/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * safe.c -- versions of standard functions that do not return on
 * error
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <errno.h>

#include "compat.h"
#include "format.h"
#include "options.h"
#include "safe.h"

#define SAFE_DIE(fmt, ...)					     \
  SAFE_CALL(mu_die, -1, fmt ": %s\n", __VA_ARGS__, strerror(errno));

/* Like SAFE_DIE, but only takes one argument. This is required
   because it is not really possible to use __VA_ARGS__ portably when
   there are no arguments. */
#define SAFE_DIE_NOARGS(msg)					\
  SAFE_CALL(mu_die, -1, msg ": %s\n" , strerror(errno));

/* Call another safe function from within one. */
#define SAFE_CALL(name, ...)				\
  _MU_SAFE_TR_NAME(name)(file, line, func, __VA_ARGS__)

/*
 * Static helper functions
 */
static int _MU_SAFE_DECL(print_err, const char *, va_list)
  __attribute__((_mu_safe_format_attr(printf, 1, 0)));

/*
 * Call mu_vdie(); see below.
 */
void _MU_SAFE_DECL(mu_die, int status, const char *fmt, ...) {
  va_list ap;

  /* Check `status' here so that we can report that it was mu_die()
     that was called erroneously, not mu_vdie(). */
  if (!status)
    SAFE_CALL(mu_die, -1, "mu_die() called with 0 status");

  va_start(ap, fmt);
  SAFE_CALL(mu_vdie, status, fmt, ap);
  va_end(ap);
}

/*
 * Call mu_vwarn(); see below.
 */
int _MU_SAFE_DECL(mu_warn, const char *fmt, ...) {
  va_list ap;
  int ret;

  va_start(ap, fmt);
  ret = SAFE_CALL(mu_vwarn, fmt, ap);
  va_end(ap);

  return ret;
}

/*
 * Print the printf-style format string `fmt' to standard error and
 * die with status `status'. If `status' is negative, abort()
 * instead. `status' must not be 0. If you want to print a warning
 * without exiting, see mu_warn() and mu_vwarn().
 */
void _MU_SAFE_DECL(mu_vdie, int status, const char *fmt, va_list ap) {
  /* If we failed to print to standard error, there's no point in
     trying again, so simply abort(). We should also abort if `status'
     is negative. */
  if (SAFE_CALL(print_err, fmt, ap) || status < 0)
    abort();
  exit(status);
}

/*
 * Like mu_vdie() (see above) except that it returns instead of
 * exiting/aborting. Returns zero if we could successfully print the
 * message, or nonzero if we couldn't.
 */
int _MU_SAFE_DECL(mu_vwarn, const char *fmt, va_list ap) {
  return SAFE_CALL(print_err, fmt, ap);
}

/*
 * The functions that follow are safe versions of the non-mu_x
 * functions.
 */

void *_MU_SAFE_DECL(mu_xmalloc, size_t size) {
  void *ptr = malloc(size);

  if (ptr)
    return ptr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate %zu bytes of memory", size);
}

void *_MU_SAFE_DECL(mu_xcalloc, size_t nmemb, size_t size) {
  void *ptr = calloc(nmemb, size);

  if (ptr)
    return ptr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate %zu bytes of memory", size);
}

void *_MU_SAFE_DECL(mu_xrealloc, void *ptr, size_t size) {
  void *res = realloc(ptr, size);

  if (res)
    return res;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to re-allocate %zu bytes of memory", size);
}

void *_MU_SAFE_DECL(mu_xreallocarray, void *ptr, size_t nmemb, size_t size) {
  void *res = reallocarray(ptr, nmemb, size);

  if (res)
    return res;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to re-allocate %zu elements, each of size %zu",
	   nmemb, size);
}

char *_MU_SAFE_DECL(mu_xstrdup, const char *str) {
  char *newstr = strdup(str);

  if (newstr)
    return newstr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate memory for a string of length %zu, \"%s\"",
	   strlen(str), str);
}

/*
 * These printf()-style functions return 'unsigned' simply to indicate
 * that the return value will never be negative. They will not,
 * however, return a value greater than INT_MAX.
 */

unsigned _MU_SAFE_DECL(mu_xasprintf, char **strp, const char *fmt, ...) {
  va_list ap;
  int ret;

  va_start(ap, fmt);
  ret = SAFE_CALL(mu_xvasprintf, strp, fmt, ap);
  va_end(ap);

  return ret;
}

unsigned _MU_SAFE_DECL(mu_xvasprintf,
		       char **strp, const char *fmt, va_list ap) {
  int ret = vasprintf(strp, fmt, ap);
  if (ret < 0)
    SAFE_DIE("failed to allocate for a string with format \"%s\"", fmt);
  return ret;
}

/*
 * These are safe versions of functions provided by Mu itself.
 */

void _MU_SAFE_DECL(mu_xformat,
		   FILE *stream, unsigned short goal, unsigned short width,
		   unsigned short *col, unsigned short indent,
		   unsigned short subindent, const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  SAFE_CALL(mu_xvformat,
	    stream, goal, width, col, indent, subindent, fmt, ap);
  va_end(ap);
}

char *_MU_SAFE_DECL(mu_xformat_string,
		    unsigned short goal, unsigned short width,
		    unsigned short *col, unsigned short indent,
		    unsigned short subindent, const char *fmt, ...) {
  va_list ap;
  char *ret;

  va_start(ap, fmt);
  ret = SAFE_CALL(mu_xvformat_string, goal, width,
		  col, indent, subindent, fmt, ap);
  va_end(ap);

  return ret;
}

void _MU_SAFE_DECL(mu_xvformat,
		   FILE *stream, unsigned short goal, unsigned short width,
		   unsigned short *col, unsigned short indent,
		   unsigned short subindent, const char *fmt, va_list ap) {
  if (mu_vformat(stream, goal, width, col, indent, subindent, fmt, ap))
    SAFE_DIE("failed to format a string with format \"%s\" to a file", fmt);
}

char *_MU_SAFE_DECL(mu_xvformat_string,
		    unsigned short goal, unsigned short width,
		    unsigned short *col, unsigned short indent,
		    unsigned short subindent, const char *fmt, va_list ap) {
  char *ret = mu_vformat_string(goal, width, col, indent, subindent, fmt, ap);
  if (!ret)
    SAFE_DIE("failed to format a string with format \"%s\" to a string", fmt);
  return ret;
}

void _MU_SAFE_DECL(mu_xformat_help,
		   FILE *stream, const char *prog_name,
		   const char *usage, const char *desc,
		   const char *notes, const MU_OPT *options) {
  if (mu_format_help(stream, prog_name, usage, desc, notes, options))
    SAFE_DIE_NOARGS("cannot print help message");
}

char *_MU_SAFE_DECL(mu_xformat_help_string,
		    const char *prog_name, const char *usage,
		    const char *desc, const char *notes,
		    const MU_OPT *options,
		    unsigned short goal, unsigned short width) {
  char *str = mu_format_help_string(prog_name, usage, desc, notes,
				    options, goal, width);
  if (!str)
    SAFE_DIE_NOARGS("cannot format help message");
  return str;
}

size_t _MU_SAFE_DECL(mu_xconcat_opts,
		     MU_OPT **res, const MU_OPT *a,
		     const MU_OPT *b, size_t extra) {
  size_t ret = mu_concat_opts(res, a, b, extra);
  if (!ret)
    SAFE_DIE("failed to concatenate two option lists: %p and %p", a, b);
  return ret;
}

/*
 * Print a message for mu_{v,}{die,warn}(). Like perl(1), if `fmt'
 * (after expansion) does not end in a newline, print information
 * about where mu_vdie() was called, and add a newline. Otherwise, the
 * message is printed verbatim.
 *
 * Returns zero if we could successfully print the message, or nonzero
 * if we couldn't.
 */
static int _MU_SAFE_DECL(print_err, const char *fmt, va_list ap) {
  char *msg;
  size_t len;
  int ret;

  /* We need to expand `fmt' into a temporary string so we can check
     if it ends in a newline. */
  len = SAFE_CALL(mu_xvasprintf, &msg, fmt, ap);
  if (!len) {
    free(msg);
    /* Generic error message in case none was supplied. */
    msg = SAFE_CALL(mu_xstrdup, "an error occurred");
    len = strlen(msg);
  }

  /* Print the message. */
  assert(len > 0);
  if (msg[len - 1] == '\n')
    ret = fputs(msg, stderr) == EOF;
  else
    ret = fprintf(stderr, "%s:%d in function %s: %s\n",
		  file, line, func, msg) < 0;

  free(msg);

  return ret;
}
